/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "data.h"
#include "bitboard.h"
#include "rodent.h"

U64 bbLineMask[4][64];
U64 attacks[4][64][64];
U64 bbPawnAttacks[2][64];
U64 bbKnightAttacks[64];
U64 bbKingAttacks[64];
U64 bbPassedMask[2][64];
U64 bbAdjacentMask[8];
int castleMask[64];

// used for a mathematical trick in FirstOne
const int bitTable[64] = {
   0,  1,  2,  7,  3, 13,  8, 19,
   4, 25, 14, 28,  9, 34, 20, 40,
   5, 17, 26, 38, 15, 46, 29, 48,
  10, 31, 35, 54, 21, 50, 41, 57,
  63,  6, 12, 18, 24, 27, 33, 39,
  16, 37, 45, 47, 30, 53, 49, 56,
  62, 11, 23, 32, 36, 44, 52, 55,
  61, 22, 43, 51, 60, 42, 59, 58
};

U64 zobPiece[12][64];
U64 zobCastle[16];
U64 zobEp[8];
int pondering;
char ponder_str[6];
int rootDepth;
int nodes;

// mobility values
static const int n_mob_mg[28] = { -16, -12, -8,  0,  4,  8,  12, 16, 16};
static const int n_mob_eg[28] = { -16, -12, -8,  0,  4,  8,  12, 16, 16};
static const int b_mob_mg[28] = { -16, -12, -8, -4, -2,  0,  2,  4,  6,  8, 10, 12, 14, 14, 16                                 };
static const int b_mob_eg[28] = { -16, -12, -8, -4, -2,  0,  2,  4,  6,  8, 10, 12, 14, 14, 16                                 };
static const int r_mob_mg[28] = {  -6,  -4, -2,  0,  2,  4,  6,  8,  9, 10, 11, 12, 13, 14, 15                                 };
static const int r_mob_eg[28] = { -12,  -8, -4,  0,  4,  8, 12, 16, 18, 20, 22, 24, 26, 28, 30                                 };
static const int q_mob_mg[28] = {  -6,  -4, -2,  0,  0,  1,  1,  1,  2,  2,  2,  3,  3,  3,  3,  4,  4,  4,  4,  5,  5,  5,  5 };
static const int q_mob_eg[28] = { -12,  -8, -4, -2,  0,  1,  2,  3,  4,  5,  6,  6,  7,  7,  8,  8,  8,  9,  9,  9, 10, 10, 10 };


// TODO: move to bb_init_masks
void sData::InitBadBishop(void) {

	 // first clear bad bishop masks
	 for (int side = 0; side < 2; side++) {
	    for (int i = 0; i < 64; i++) {  
			 Data.bbBadBishopMasks [side][i] = 0ULL;
		 }
	 }

	 Data.bbBadBishopMasks[WHITE][F1] = SqBb(E2);
	 Data.bbBadBishopMasks[BLACK][F8] = SqBb(E7);

	 Data.bbBadBishopMasks[WHITE][C1] = SqBb(D2);
	 Data.bbBadBishopMasks[BLACK][C8] = SqBb(D7);

	 Data.bbBadBishopMasks[WHITE][A2] = SqBb(B3) | SqBb(C4);
	 Data.bbBadBishopMasks[BLACK][A7] = SqBb(B6) | SqBb(C5);

	 Data.bbBadBishopMasks[WHITE][B2] = SqBb(C3);
	 Data.bbBadBishopMasks[BLACK][B7] = SqBb(C6);

 	 Data.bbBadBishopMasks[WHITE][C2] = SqBb(D3) | SqBb(E4);
	 Data.bbBadBishopMasks[BLACK][C7] = SqBb(D6) | SqBb(E5);

 	 Data.bbBadBishopMasks[WHITE][D2] = SqBb(E3);
	 Data.bbBadBishopMasks[BLACK][D7] = SqBb(E6);

 	 Data.bbBadBishopMasks[WHITE][E2] = SqBb(D3);
	 Data.bbBadBishopMasks[BLACK][E7] = SqBb(D6);

 	 Data.bbBadBishopMasks[WHITE][F2] = SqBb(E3) | SqBb(D4);
	 Data.bbBadBishopMasks[BLACK][F7] = SqBb(E6) | SqBb(D5);

	 Data.bbBadBishopMasks[WHITE][G2] = SqBb(F3);
	 Data.bbBadBishopMasks[BLACK][G7] = SqBb(F6);

 	 Data.bbBadBishopMasks[WHITE][H2] = SqBb(G3) | SqBb(F4);
	 Data.bbBadBishopMasks[BLACK][H7] = SqBb(G6) | SqBb(F5);

	 Data.bbBadBishopMasks[WHITE][A3] = SqBb(B4) | SqBb(C5);
	 Data.bbBadBishopMasks[BLACK][A6] = SqBb(B5) | SqBb(C4);

	 Data.bbBadBishopMasks[WHITE][B3] = SqBb(C4) | SqBb(D5);
	 Data.bbBadBishopMasks[BLACK][B6] = SqBb(C5) | SqBb(D4);

	 Data.bbBadBishopMasks[WHITE][C3] = SqBb(D4);
	 Data.bbBadBishopMasks[BLACK][C6] = SqBb(D5);

	 Data.bbBadBishopMasks[WHITE][D3] = SqBb(E4);
	 Data.bbBadBishopMasks[BLACK][D6] = SqBb(E5);

	 Data.bbBadBishopMasks[WHITE][E3] = SqBb(D4);
	 Data.bbBadBishopMasks[BLACK][E6] = SqBb(D5);

	 Data.bbBadBishopMasks[WHITE][F3] = SqBb(E4);
	 Data.bbBadBishopMasks[BLACK][F6] = SqBb(E5);

	 Data.bbBadBishopMasks[WHITE][G3] = SqBb(F4) | SqBb(E5);
	 Data.bbBadBishopMasks[BLACK][G6] = SqBb(F5) | SqBb(E4);

	 Data.bbBadBishopMasks[WHITE][H3] = SqBb(G4) | SqBb(F5);
	 Data.bbBadBishopMasks[BLACK][H6] = SqBb(G5) | SqBb(F4);

	 Data.bbBadBishopMasks[WHITE][C4] = SqBb(D5);
	 Data.bbBadBishopMasks[BLACK][C5] = SqBb(D4);

	 Data.bbBadBishopMasks[WHITE][F4] = SqBb(E5);
	 Data.bbBadBishopMasks[BLACK][F5] = SqBb(E4);
}

void sData::InitMaterialValues(void) {
	//             name,  val, pha, att, wood
	 InitSinglePiece( P,  0,    0,   0,   0);  // pawn material is evaluated in eval.c
	 InitSinglePiece( N,  325,  1,  20,   1);
	 InitSinglePiece( B,  335,  1,  20,   1);
	 InitSinglePiece( R,  500,  2,  40,   1);
	 InitSinglePiece( Q,  975,  4,  80,   1);
	 InitSinglePiece(NO_TP, 0,  0,   0,   0);
};

void sData::InitSinglePiece(int pc, int mat, int phase, int att, int wood) {
     matValue[pc]      = mat;   
	 phaseValue[pc]    = phase;     
	 attMultiplier[pc] = att;      
	 woodCount[pc]     = wood;
}

void sData::InitMobBonus(void) {
	 for (int i = 0; i < 28; i++) {
		 mobBonusMg[N][i] = n_mob_mg[i];
		 mobBonusEg[N][i] = n_mob_eg[i];
		 mobBonusMg[B][i] = b_mob_mg[i];
		 mobBonusEg[B][i] = b_mob_eg[i];
		 mobBonusMg[R][i] = r_mob_mg[i];
		 mobBonusEg[R][i] = r_mob_eg[i];
		 mobBonusMg[Q][i] = q_mob_mg[i];
		 mobBonusEg[Q][i] = q_mob_eg[i];
	 }
}

void sData::InitCastleMask(void) {
  for (int i = 0; i < 64; i++) castleMask[i] = 15;
  castleMask[A1] = 13;
  castleMask[E1] = 12;
  castleMask[H1] = 14;
  castleMask[A8] = 7;
  castleMask[E8] = 3;
  castleMask[H8] = 11;
}

void sData::InitSearchData(void) {

     // History of moveIsLate:lmrStep tuning:
	 // 10:5 best, 8:4 best, 8:5 fails, 6:4 best, 5:4 fails)

     aspiration       = 30;
	 useNull          = 1;
	 minimalNullDepth = ONE_PLY;
	 useLmr           = 1;
	 minimalLmrDepth  = 2 * ONE_PLY;
	 moveIsLate       = 6;
	 lmrStep          = 4; 
	 useReSearch      = 1;
	 useFutility      = 1;
	 futilityBase     = 100;
	 futilityStep     = 20;
	 futilityDepth    = 4;
	 useDeltaPruning  = 1;
	 deltaMargin      = 150;
	 goodCaptMargin   = 30;   // BxN is OK, RxB isn't
	 lazyMargin       = 180;
	 verbose          = 0; // no additional display
#ifdef USE_WEAKENING
	 useWeakening     = 1;
	 elo              = MAX_ELO;
#endif
}

void sData::InitDistanceBonus(void) 
{
    for (int i = 0; i < 64; ++i) {
         for (int j = 0; j < 64; ++j) {
         distance[i][j] = 14 - ( Abs( Rank(i) - Rank(j) ) + Abs( File(i) - File(j) ) );
         }
     }
}

void sData::InitEvalVars(void) 
{
	 ownMobility   = 100;
	 oppMobility   = 100;
	 ownAttack     = 100;
	 oppAttack     = 100;
	 bishopPair    = 30;
	 knightPair    = -10;
	 rookPair      = -10;
	 doubledPawnMg = -20;
     doubledPawnEg = -10;
	 rookOpenMg    = 10;
     rookOpenEg    = 10;
	 rookSemiOpenMg = 5;
	 rookSemiOpenEg = 5;
	 rookOpenAttack = 10;
	 rookSemiOpenAttack = 5;
	 rookSeventhMg  = 20;
	 rookSeventhEg  = 20;
	 shieldBase     = 5; // was 5 
	 shieldSecond   = 5; // was 5
	 shieldThird    = 2; // was 2 LOWER IT 
	 shieldMissing  = -15; // INCREASE IT
	 pawnIsolatedOnOpen = -10;
	 pawnBackwardOnOpen = -10;
}

void sData::InitAsymmetric(int side) {
     mobSidePercentage[side]      = ownMobility;
     mobSidePercentage[Opp(side)] = oppMobility;
     attSidePercentage[side]      = ownAttack;
     attSidePercentage[Opp(side)] = oppAttack;
}