/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include "bitboard.h"
#include "bitmasks.h"
#include "data.h"
#include "rodent.h"
#include "eval.h"

void sEvaluator::InitStatic(void) 
{
  mgScore            = 0;   egScore            = 0;  // clear midgame/endgame score component
//  mgPassers[WHITE]   = 0;   mgPassers[BLACK]   = 0;  // clear passer midgame scores
//  egPassers[WHITE]   = 0;   egPassers[BLACK]   = 0;  // clear passer endgame scores
  pawnScoreMg[WHITE] = 0;   pawnScoreMg[BLACK] = 0;  // clear midgame pawn scores
  pawnScoreEg[WHITE] = 0;   pawnScoreEg[BLACK] = 0;  // clear endgame pawn scores	 
  degradation        = 64;                           // no scaling for drawish endgames
}


void sEvaluator::InitDynamic(void) {
  woodCount[WHITE]   = 0;   woodCount[BLACK]   = 0; 
  attCount[WHITE]    = 0;   attCount[BLACK]    = 0; 
  attScore[WHITE]    = 0;   attScore[BLACK]    = 0;  // clear attack scores
  mgMisc[WHITE]      = 0;   mgMisc[BLACK]      = 0;  // clear miscelanneous midgame scores
  egMisc[WHITE]      = 0;   egMisc[BLACK]      = 0;  // clear miscelanneous endgame scores
  mgMobility[WHITE]  = 0;   mgMobility[BLACK]  = 0;  // clear midgame mobility
  egMobility[WHITE]  = 0;   egMobility[BLACK]  = 0;  // clear endgame mobility	 

}

void sEvaluator::InitPawnData(sPosition *p) {
     bbPawnControl[WHITE]    = GetWPControl( bbPc(p, WHITE, P) );
	 bbPawnControl[BLACK]    = GetBPControl( bbPc(p, BLACK, P) );
	 bbPawnCanControl[WHITE] = FillNorth( bbPawnControl[WHITE] );
	 bbPawnCanControl[BLACK] = FillSouth( bbPawnControl[BLACK] );
}

void sEvaluator::SetScaleFactor(sPosition *p) 
{
  mgFact = Min( p->phase, 24); // normalize for opening material
  egFact = 24 - mgFact;
}

// Interpolate between mgScore and egScore, depending on remaining material 
int sEvaluator::Interpolate(void) 
{
  return ( (mgFact * mgScore ) / 24 ) + ( (egFact * egScore ) / 24 );
}

void sEvaluator::SetKingZones(sPosition *p) {
  // set bitboard of squares around enemy king
  bbKingZone[WHITE]    = bbKingAttacks[KingSq(p, WHITE ^ 1) ]; 
  bbKingZone[BLACK]    = bbKingAttacks[KingSq(p, BLACK ^ 1) ]; 

  // set zone for evaluationg king's pawn shield, using shifted version of bbKingZone
  bbShieldZone[WHITE]  = bbKingZone[WHITE] << 8;
  bbShieldZone[BLACK]  = bbKingZone[BLACK] >> 8;

  // enlarge king attack zone by squares in front of the ring, facing enemy position
  bbKingZone[WHITE]   |= bbShieldZone[WHITE];
  bbKingZone[BLACK]   |= bbShieldZone[BLACK];

  // set squares from which a king can be checked 
  U64 bbOccupied = OccBb(p);
  kingStraightChecks[WHITE] = RAttacks(bbOccupied, KingSq(p, WHITE) );  
  kingStraightChecks[BLACK] = RAttacks(bbOccupied, KingSq(p, BLACK) );  
  kingDiagChecks[WHITE]     = BAttacks(bbOccupied, KingSq(p, WHITE) );  
  kingDiagChecks[BLACK]     = BAttacks(bbOccupied, KingSq(p, BLACK) );  
}

void sEvaluator::EvalKings(sPosition *p) 
{
  // king safety scaling is based on Fruit
  const int attMult[8] = {0, 10, 55, 75, 83, 86, 92, 100};

  U64 bbFrontSpan;
  U64 bbSideSpan;
  U64 bbWhiteShield = bbShieldZone[WHITE] & bbPc( p, WHITE, P );
  U64 bbBlackShield = bbShieldZone[BLACK] & bbPc( p, BLACK, P );

  // general shield bonus
  mgScore += Data.shieldBase * PopCntSparse(bbWhiteShield);  
  mgScore -= Data.shieldBase * PopCntSparse(bbBlackShield);

  // additional bonus for pawns on the 2nd/3rd rank
  mgScore += Data.shieldSecond * PopCntSparse(bbWhiteShield & bbRANK_2);
  mgScore -= Data.shieldSecond * PopCntSparse(bbBlackShield & bbRANK_7);
  mgScore += Data.shieldThird  * PopCntSparse(bbWhiteShield & bbRANK_2);
  mgScore -= Data.shieldThird  * PopCntSparse(bbBlackShield & bbRANK_7);

  // penalty for missing shielding pawns
  bbFrontSpan = GetFrontSpan(SqBb(KingSq(p, WHITE )), WHITE );
  if (!(bbFrontSpan & bbPc(p,WHITE,P) ) ) mgScore  += Data.shieldMissing;

  bbSideSpan = ShiftEast(bbFrontSpan);
  if (bbSideSpan) {
     if (!(bbSideSpan & bbPc(p,WHITE,P) ) ) mgScore  += Data.shieldMissing;
  }
  bbSideSpan = ShiftWest(bbFrontSpan);
  if (bbSideSpan) {
     if (!(bbSideSpan & bbPc(p,WHITE,P) ) ) mgScore  += Data.shieldMissing;
  }

  bbFrontSpan = GetFrontSpan(SqBb(KingSq(p, BLACK )), BLACK );
  if (!(bbFrontSpan & bbPc(p,BLACK,P) ) ) mgScore  -= Data.shieldMissing;

  bbSideSpan = ShiftEast(bbFrontSpan);
  if (bbSideSpan) {
     if (!(bbSideSpan & bbPc(p,BLACK,P) ) ) mgScore  -= Data.shieldMissing;
  }

  bbSideSpan = ShiftWest(bbFrontSpan);
  if (bbSideSpan) {
     if (!(bbSideSpan & bbPc(p,BLACK,P) ) ) mgScore  -= Data.shieldMissing;
  }

  if (woodCount[WHITE] > 7) woodCount[WHITE] = 7; // normalize
      attScore[WHITE] = ( ( attCount[WHITE] * attMult[woodCount[WHITE] ] ) / 100 ); 

  if (woodCount[BLACK] > 7) woodCount[BLACK] = 7; // normalize
      attScore[BLACK] = ( ( attCount[BLACK] * attMult[woodCount[BLACK] ] ) / 100 ); 
}

// full evaluation function
int sEvaluator::Return(sPosition *p, int alpha, int beta)
{
  int score = GetMaterialScore(p) + AdjustMaterial(p);

  InitStatic();
  SetScaleFactor(p);
  EvalPawns(p);
  //EvalPassers(p, WHITE);
  //EvalPassers(p, BLACK);

  score += EvalTrappedKnight(p);                 
  score += EvalTrappedBishop(p);                 
  score += EvalTrappedRook(p);  

  // TEMPO
  if (p->side == WHITE) mgScore += 5;
  else                  mgScore -= 5;

  mgScore += (p->pstMg[WHITE] - p->pstMg[BLACK]);
  egScore += (p->pstEg[WHITE] - p->pstEg[BLACK]);
  //mgScore += mgPassers[WHITE] - mgPassers[BLACK];
  //egScore += egPassers[WHITE] - egPassers[BLACK];
  
  int temp_score = score + Interpolate();

  // lazy evaluation - we avoid costly procedures
  // if score seems already very high/very low
    if (temp_score > alpha - Data.lazyMargin 
    &&  temp_score < beta +  Data.lazyMargin) 
  {
	  SetKingZones(p);
	  InitPawnData(p);
	  InitDynamic();

      ScoreN(p, &attCount[WHITE], &woodCount[WHITE], WHITE);
      ScoreN(p, &attCount[BLACK], &woodCount[BLACK], BLACK);
	  ScoreB(p, &attCount[WHITE], &woodCount[WHITE], WHITE);
      ScoreB(p, &attCount[BLACK], &woodCount[BLACK], BLACK);
      ScoreR(p, &attCount[WHITE], &woodCount[WHITE], WHITE);
	  ScoreR(p, &attCount[BLACK], &woodCount[BLACK], BLACK);
      ScoreQ(p, &attCount[WHITE], &woodCount[WHITE], WHITE);
      ScoreQ(p, &attCount[BLACK], &woodCount[BLACK], BLACK);

	  EvalKings(p);
      
	  // ASYMMETRIC MOBILITY AND ATTACK SCALING
	  ScaleValue(&mgMobility[WHITE], Data.mobSidePercentage[WHITE]);
	  ScaleValue(&mgMobility[BLACK], Data.mobSidePercentage[BLACK]);
  	  ScaleValue(&egMobility[WHITE], Data.mobSidePercentage[WHITE]);
	  ScaleValue(&egMobility[BLACK], Data.mobSidePercentage[BLACK]);
	  ScaleValue(&attScore[WHITE]  , Data.attSidePercentage[WHITE]);
	  ScaleValue(&attScore[BLACK]  , Data.attSidePercentage[BLACK]);

	  // MERGING SCORE
	  mgScore += ( mgMobility[WHITE] - mgMobility[BLACK] );
	  egScore += ( egMobility[WHITE] - egMobility[BLACK] );
	  mgScore += ( mgMisc[WHITE]     - mgMisc[BLACK]     );
	  egScore += ( egMisc[WHITE]     - egMisc[BLACK]     );
      score   += Interpolate();    // merge middlegame and endgame scores
	  score   += ( attScore[WHITE]  - attScore[BLACK] );

      // "COSMETIC" CODE PREVENTING UGLY OPENING MOVES
	  if  ( ( bbPc(p, WHITE, N) & SqBb(C3) )
	  &&  ( bbPc(p, WHITE, P) & SqBb(C2) )
	  &&  ( bbPc(p, WHITE, P) & SqBb(D4) )
	  &&  !( bbPc(p, WHITE, P) & SqBb(E4) )
	  ) score -= 15;

      if  ( ( bbPc(p, BLACK, N) & SqBb(C6) )
	  &&  ( bbPc(p, BLACK, P) & SqBb(C7) )
	  &&  ( bbPc(p, BLACK, P) & SqBb(D5) )
	  &&  !( bbPc(p, BLACK, P) & SqBb(E5) )
	  ) score += 15;

  }
  else score = temp_score; 

  // incentive for the final simplification (so that KRR vs KR is won)
  if ( p->pcCount[WHITE][P] == 0 && p->pieceMat[WHITE] == 0 ) score -= 50;
  if ( p->pcCount[BLACK][P] == 0 && p->pieceMat[BLACK] == 0 ) score += 50;

  if (score == 0) score = 1; // so that 0 returned by search means draw

  // decrease score for drawish endgames
  SetScoreDegradation(p, score);
  score *= degradation;
  score /= 64;

  // normalize_score
  score = Normalize(score, MAX_EVAL);

#ifdef USE_WEAKENING
  if (Data.elo < MAX_ELO && Data.useWeakening) {
	  int randomFactor = ( MAX_ELO - Data.elo ) / 15;
	  int randomMod = (randomFactor / 2) - (p->hashKey % randomFactor);
	  score += randomMod;
  }
#endif

  // return score relative to the side to move
  return p->side == WHITE ? score : -score;
}

int sEvaluator::ReturnMaterial(sPosition *p) 
{
    int score = GetMaterialScore(p);
	return p->side == WHITE ? score : -score;
}

// fast evaluation function (material, pst, pawn structure)
int sEvaluator::ReturnFast(sPosition *p)
{
  int score = GetMaterialScore(p) + AdjustMaterial(p);

  InitStatic();
  SetScaleFactor(p);
  EvalPawns(p);
  //EvalPassers(p, WHITE);
  //EvalPassers(p, BLACK);

  // TEMPO
  if (p->side == WHITE) mgScore += 5;
  else                  mgScore -= 5;

  mgScore += (p->pstMg[WHITE] - p->pstMg[BLACK]);
  egScore += (p->pstEg[WHITE] - p->pstEg[BLACK]);
  //mgScore += mgPassers[WHITE] - mgPassers[BLACK];
  //egScore += egPassers[WHITE] - egPassers[BLACK];
  
  score += Interpolate();

  // incentive for the final simplification (so that for example KRR vs KR is won)
  if ( p->pcCount[WHITE][P] == 0 && p->pieceMat[WHITE] == 0 ) score -= 50;
  if ( p->pcCount[BLACK][P] == 0 && p->pieceMat[BLACK] == 0 ) score += 50;

  // decrease score for drawish endgames
  SetScoreDegradation(p, score);
  score *= degradation;
  score /= 64;

  // normalize_score
  score = Normalize(score, MAX_EVAL);
  if (score == 0) score = 1; // so that 0 returned by search means draw

  // return score relative to the side to move
  return p->side == WHITE ? score : -score;
}

int sEvaluator::Normalize(int val, int limit) {
	if (val > limit)        return limit;
	else if (-val < -limit) return -limit;
    return val;
}

void sEvaluator::ScaleValue(int * value, int factor) {
     *value *= factor;
	 *value /= 100;
 }
