/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "rodent.h"
#include "data.h"
#include "bitboard.h"
#include "bitmasks.h"
#include "eval.h"

// Functions responsible for scaling down the evaluation score
// in case of drawish endgames.

void sEvaluator::SetScoreDegradation(sPosition *p, int score) 
{
  // no score degradation in the middlegame or in the endgame with many pawns (speedup)
  if ( p->pieceMat[WHITE] > 1400
  ||   p->pieceMat[BLACK] > 1400 
  ||   p->pcCount[WHITE][P] > 2 
  ||   p->pcCount[BLACK][P] > 2 ) return;

  // unwinnable endgames - stronger side has at most bare minor piece
  if (p->pcCount[WHITE][P] == 0 
  && p->pieceMat[WHITE] < 400 
  && score > 0) { degradation = 0; return; };
  
  if (p->pcCount[BLACK][P] == 0 
  && p->pieceMat[BLACK] < 400 
  && score < 0) { degradation = 0; return; };

  // no pawns of either color on the board
  if (p->pcCount[WHITE][P] == 0 
  &&  p->pcCount[BLACK][P] == 0) {

	  // unwinnable endgame: KNN vs K
	  if ( MaterialNN(p, WHITE) && score < 0 ) { degradation = 0; return; };
	  if ( MaterialNN(p, BLACK) && score > 0 ) { degradation = 0; return; };
	  // TODO: what about KNN vs KN?

	  // low and almost equal material, except KBB vs KN:     1/8
	  if ( MaterialBN(p, WHITE) && ( MaterialMinor (p, BLACK) || MaterialRook(p, BLACK) ) ) { degradation = 8; return; };
	  if ( MaterialBN(p, BLACK) && ( MaterialMinor (p, WHITE) || MaterialRook(p, WHITE) ) ) { degradation = 8; return; };
	  if ( MaterialBB(p, WHITE) && ( MaterialBishop(p, BLACK) || MaterialRook(p, BLACK) ) ) { degradation = 8; return; };
	  if ( MaterialBB(p, BLACK) && ( MaterialBishop(p, WHITE) || MaterialRook(p, WHITE) ) ) { degradation = 8; return; };
  } // no pawns

  // it's hard to win with a bare rook against a minor/minor + pawn
  if (p->pcCount[WHITE][P] == 0 
  && p->pieceMat[WHITE] == Data.matValue[R] 
  && p->pieceMat[BLACK] > 0 
  && score > 0) { degradation = 32; return; }; // 1/2

  if (p->pcCount[BLACK][P] == 0 
  && p->pieceMat[BLACK] == Data.matValue[R] 
  && p->pieceMat[WHITE] > 0 
  && score < 0) { degradation = 32; return; }; // 1/2

  // it's hard to win with a rook and a minor against a rook/rook + pawns
  if (p->pcCount[WHITE][P] == 0 
  && MaterialRookMinor(p, WHITE) 
  && MaterialRook(p, BLACK)
  && score > 0) { degradation = 32; return; }; // 1/2

  if (p->pcCount[BLACK][P] == 0 
  && MaterialRookMinor(p, BLACK) 
  && MaterialRook(p, WHITE)
  && score < 0) { degradation = 32; return; }; // 1/2

  // it's hard to win with a queen and a minor against a queen/queen + pawns
  if (p->pcCount[WHITE][P] == 0 
  && MaterialQueenMinor(p, WHITE) 
  && MaterialQueen(p, BLACK)
  && score > 0) { degradation = 16; return; }; // 1/2

  if (p->pcCount[BLACK][P] == 0 
  && MaterialQueenMinor(p, BLACK) 
  && MaterialQueen(p, WHITE)
  && score < 0) { degradation = 32; return; }; // 1/2

  // decrease score in pure opposite bishops ending
  if (MaterialBishop(p, WHITE) 
  && MaterialBishop(p, BLACK)
  && BishopsAreDifferent(p)
  ) { degradation = 32; return; }; // 1/2

  // KBP vs Km is drawn when defending king stands on pawn's path 
  // and cannot be driven out by a Bishop
  if (MaterialBishop(p, WHITE)
  &&  MaterialMinor(p, BLACK)
  &&  p->pcCount[WHITE][P] == 1
  &&  p->pcCount[BLACK][P] == 0
  &&  ( SqBb(p->kingSquare[BLACK]) & GetFrontSpan( bbPc(p, WHITE, P), WHITE ) )
  &&  NotOnBishColor(p, WHITE, p->kingSquare[BLACK])
  ) { degradation = 0; return; }; // draw score

  if (MaterialBishop(p, BLACK)
  &&  MaterialMinor(p, WHITE)
  &&  p->pcCount[BLACK][P] == 1
  &&  p->pcCount[WHITE][P] == 0
  &&  ( SqBb(p->kingSquare[WHITE]) & GetFrontSpan( bbPc(p, BLACK, P), BLACK ) )
  &&  NotOnBishColor(p, BLACK, p->kingSquare[WHITE])
  ) { degradation = 0; return; }; // draw score

  // good defensive position in KRK
  if (MaterialRook(p, WHITE)
  &&  MaterialRook(p, BLACK)
  &&  p->pcCount[WHITE][P] == 1
  &&  p->pcCount[BLACK][P] == 0
  &&  ( SqBb(p->kingSquare[BLACK]) & GetFrontSpan( bbPc(p, WHITE, P), WHITE ) )
  &&  NotOnBishColor(p, WHITE, p->kingSquare[BLACK])
  ) { degradation = 32; return; }; // draw score

  if (MaterialRook(p, BLACK)
  &&  MaterialRook(p, WHITE)
  &&  p->pcCount[BLACK][P] == 1
  &&  p->pcCount[WHITE][P] == 0
  &&  ( SqBb(p->kingSquare[WHITE]) & GetFrontSpan( bbPc(p, BLACK, P), BLACK ) )
  &&  NotOnBishColor(p, BLACK, p->kingSquare[WHITE])
  ) { degradation = 32; return; }; // draw score

}

int BishopsAreDifferent(sPosition * p) {
	if ( ( bbWhiteSq & bbPc(p, WHITE, B) ) && ( bbBlackSq & bbPc(p, BLACK, B) ) ) return 1;
	if ( ( bbBlackSq & bbPc(p, WHITE, B) ) && ( bbWhiteSq & bbPc(p, BLACK, B) ) ) return 1;
    return 0;
}

int NotOnBishColor(sPosition * p, int bishSide, int sq) {
    if ( ( ( bbWhiteSq & bbPc(p, bishSide, B) ) == 0 )
    && ( SqBb(sq) & bbWhiteSq) ) return 1;

    if ( ( ( bbBlackSq & bbPc(p, bishSide, B) ) == 0 )
    && ( SqBb(sq) & bbBlackSq) ) return 1;

    return 0;
}

int MaterialKnight(sPosition *p, int side) {
	return ( p->pieceMat[side] == Data.matValue[N] );
}

int MaterialBishop(sPosition *p, int side) {
	return ( p->pieceMat[side] == Data.matValue[B] );
}

int MaterialMinor(sPosition *p, int side) {
   return (    (p->pieceMat[side] ==  Data.matValue[B]) 
	        || (p->pieceMat[side] ==  Data.matValue[N]) );
}

int MaterialRook(sPosition *p, int side) {
	return ( p->pieceMat[side] == Data.matValue[R] );
}

int MaterialRookMinor(sPosition *p, int side) {
    return (    (p->pieceMat[side] == Data.matValue[R] + Data.matValue[B]) 
	         || (p->pieceMat[side] == Data.matValue[R] + Data.matValue[N]) );
}

int MaterialQueen(sPosition *p, int side) {
	return ( p->pieceMat[side] == Data.matValue[Q] );
}

int MaterialQueenMinor(sPosition *p, int side) {
    return (    (p->pieceMat[side] == Data.matValue[Q] + Data.matValue[B]) 
	         || (p->pieceMat[side] == Data.matValue[Q] + Data.matValue[N]) );
}

int MaterialNN(sPosition *p, int side) {
    return ( p->pieceMat[side] == 2*Data.matValue[N] );
}

int MaterialBB(sPosition *p, int side) {
    return ( p->pieceMat[side] == 2*Data.matValue[B] );
}

int MaterialBN(sPosition *p, int side) {
	return ( p->pieceMat[side] == Data.matValue[B] + Data.matValue[N] );
}