/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "rodent.h"
#include "data.h"
#include "eval.h"

// TODO: move to data struct
static const int pawnMat[9] = {-40, 90, 180, 270, 360, 450, 540, 630, 720};

/**********************************************************************
/  Crafty-like material imbalance table indexed by difference         /
/  in major piece material (R) and minor piece material (n)           /
**********************************************************************/

#define A 80 // advantage in both major and minor pieces

static const int imbalance[9][9] = {
/* n=-4  n=-3  n=-2  n=-1  n=0   n=+1  n=+2  n=+3  n=+4 */
  {  -A,   -A,   -A,   -A,  -50,    0,    0,    0,    0 }, // R = -4 
  {  -A,   -A,   -A,   -A,  -50,    0,    0,    0,    0 }, // R = -3 
  {  -A,   -A,   -A,   -A,  -50,    0,    0,    0,    0 }, // R = -2 
  {  -A,   -A,   -A,   -A,  -50,  -10,   60,    0,    0 }, // R = -1 
  { -40,  -40,  -40,  -40,    0,   40,   40,   40,   40 }, // R =  0 
  {   0,    0,  -60,   10,   50,    A,    A,    A,    A }, // R = +1 
  {   0,    0,    0,    0,   50,    A,    A,    A,    A }, // R = +2 
  {   0,    0,    0,    0,   50,    A,    A,    A,    A }, // R = +3 
  {   0,    0,    0,    0,   50,    A,    A,    A,    A }  // R = +4 
};

int sEvaluator::GetMaterialScore(sPosition *p) 
{
  int score = p->pieceMat[WHITE] - p->pieceMat[BLACK]; // piece material
  score += pawnMat[p->pcCount[WHITE][P]];
  score -= pawnMat[p->pcCount[BLACK][P]];
  return score;
}

int sEvaluator::AdjustMaterial(sPosition *p) 
{
  int result = 0;
  int x, y;
  int minorBalance = p->pcCount[WHITE][N] - p->pcCount[BLACK][N] + p->pcCount[WHITE][B]   - p->pcCount[BLACK][B];
  int majorBalance = p->pcCount[WHITE][R] - p->pcCount[BLACK][R] + 2*p->pcCount[WHITE][Q] - 2*p->pcCount[BLACK][Q];

  // piece pairs
  if ( p->pcCount[WHITE][B] > 1) result += Data.bishopPair;
  if ( p->pcCount[BLACK][B] > 1) result -= Data.bishopPair;

  if ( p->pcCount[WHITE][N] > 1) result += Data.knightPair;
  if ( p->pcCount[BLACK][N] > 1) result -= Data.knightPair;

  if ( p->pcCount[WHITE][B] > 1) result += Data.rookPair;
  if ( p->pcCount[BLACK][B] > 1) result -= Data.rookPair;

  // material imbalance table
  x = Max( majorBalance + 4, 0 );
  if (x > 8) x = 8;

  y = Max(minorBalance + 4, 0);
  if (y > 8) y = 8;

  result += imbalance [x] [y];

  // checkmate help
  // TODO: special cases KBNK, KRKN
  if (p->pcCount[WHITE][P] == 0
  &&  p->pcCount[BLACK][P] == 0) {

      if (result > 100
	  && p->pieceMat[BLACK] < 600)
	  result += (40 - Data.pstEg[BLACK][K][p->kingSquare[BLACK]] + Data.distance[p->kingSquare[WHITE]] [p->kingSquare[BLACK]]);

      if (result < 100
	  && p->pieceMat[WHITE] < 600)
	  result -= (40 - Data.pstEg[WHITE][K][p->kingSquare[WHITE]] + Data.distance[p->kingSquare[WHITE]] [p->kingSquare[BLACK]]);
  }

  return result;
}