/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  This file contains functions evaluating the pieces; please note
  that trapped pieces evaluation is coded in eval_trapped.c
*/

#include <stdio.h>
#include "rodent.h"
#include "bitboard.h"
#include "bitmasks.h"
#include "data.h"
#include "gencache.h"
#include "eval.h"

#define QUEEN_CAN_CHECK 10
#define ROOK_CAN_CHECK   5
#define KNIGHT_IN_HOLE   4
#define BISHOP_IN_HOLE   4

/*
Knight evaluation:

- outposts (placed in holes of enemy pawn structure, further bonus if defended)
- mobility (excluding squares attacked by enemy pawns)
- king attacks
*/

void sEvaluator::ScoreN(sPosition *p, int *attCount, int *woodCount, int side) 
{
  int sq;
  U64 bbControl, bbAttZone;
  U64 bbPieces = bbPc(p, side, N);

  while (bbPieces) {
    sq = FirstOne(bbPieces);
	bbControl = bbKnightAttacks[sq] & ~p->bbCl[side];
	bbAttZone = bbControl & bbKingZone[side];

	// our knight sits in a hole of enemy pawn structure
	if ( SqBb(sq) & ~bbPawnCanControl[Opp(side)] ) {
		AddMiscOne(side, KNIGHT_IN_HOLE);

	   // additional bonus if defended by a pawn
	   if ( SqBb(sq) & bbPawnControl[side] ) 
		   AddMiscOne(side, Data.outpostKnight[side][sq] );
	}

	// attack
	if (bbAttZone && p->pcCount[side][Q] ) {
	   *attCount  += Data.attMultiplier[N] * PopCntSparse(bbAttZone);
	   *woodCount += Data.woodCount[N]; 
	}

    // exclude squares controlled by enemy pawns and evaluate mobility
	bbControl &= ~bbPawnControl[Opp(side)];
	AddMobility(N, side, PopCnt15(bbControl) );
    
	bbPieces &= bbPieces - 1;
  }
}

/*
Bishop evaluation:

- outposts (placed in holes of enemy pawn structure, further bonus if defended)
- bad bishop (= own pawns on certain squares are blocking it)
- mobility (including attacks through own queen)
- check threats (DISABLED)
- king attacks (including attacks through own queen)
*/

void sEvaluator::ScoreB(sPosition *p, int *attCount, int *woodCount, int side) 
{
  int sq;
  U64 bbControl, bbAttZone;
  U64 bbPieces      = bbPc(p, side, B);
  U64 bbOccupied    = OccBb(p) ^ bbPc(p, side, Q); // we accept mobility through own queen

  while (bbPieces) {
    sq = FirstOne(bbPieces);
	bbControl = GenCache.GetBishMob(bbOccupied, sq);

	// our bishop sits in a hole of enemy pawn structure
	if ( SqBb(sq) & ~bbPawnCanControl[Opp(side)] ) {
		AddMiscOne(side, BISHOP_IN_HOLE);

	    // additional bonus if defended by a pawn
		if ( SqBb(sq) & bbPawnControl[side] ) 
           AddMiscOne(side, Data.outpostBishop[side][sq] );
	}

    // check threats - PLAYS WORSE WITH IT
	/*if (bbControl & ( kingDiagChecks[Opp(side)] ) )
		*attCount += 2; */


    /* if we can attack zone around enemy king from the current square, 
	/  test this possibility on the actual board. */
	if (bbBCanAttack[sq] [KingSq(p, side ^ 1) ] 
	&& (bbControl & bbKingZone[side])  ) {

	   // count king attacks
       bbAttZone = bbControl & bbKingZone[side];
	
	   if (bbAttZone && p->pcCount[side][Q] ) {
		   *attCount  += Data.attMultiplier[B] * PopCntSparse(bbAttZone);
		   *woodCount += Data.woodCount[B]; 
	   }
   }

    // exclude squares controlled by enemy pawns and evaluate mobility
	bbControl &= ~bbPawnControl[Opp(side)];
	AddMobility(B, side, PopCnt15(bbControl) );

	// penalize bishop blocked by own pawns
	if ( Data.bbBadBishopMasks[side][sq] & bbPc(p, side, P) )
       AddMiscOne(side, Data.badBishopPenalty[side][sq]);
    
	bbPieces &= bbPieces - 1;
  }
}

/*
Rook evaluation:

- outpost (defended by a pawn, sitting in the howle of enemy pawn structure)
- half/open and open files (including king attack bonus)
- 7th rank
- mobility (including attacks through own major pieces)
- check threats
- king attacks (including attacks through the other rook / through a queen)
*/

void sEvaluator::ScoreR(sPosition *p, int *attCount, int *woodCount, int side) 
{
  int sq;
  U64 bbControl, bbAttZone;
  U64 bbPieces   = bbPc(p, side, R);
  U64 bbOccupied = OccBb(p) ^ bbPc(p, side, Q) ^ bbPc(p, side, R);
  const U64 bbSeventh[2] = { bbRANK_7, bbRANK_2};
  const U64 bbEighth[2]  = { bbRANK_8, bbRANK_1};

  while (bbPieces) {
    sq = FirstOne(bbPieces);

	// our rook sits in a hole of enemy pawn structure
	if ( SqBb(sq) & ~bbPawnCanControl[Opp(side)] ) {

	    // our rook is defended by a pawn
		if ( SqBb(sq) & bbPawnControl[side] ) {
		   AddMiscOne(side, Data.outpostRook[side][sq]);
	    }
	}

	bbControl = GenCache.GetRookMob(bbOccupied, sq);

	// evaluate rook on an open file
	U64 bbFrontSpan = GetFrontSpan(SqBb(sq), side );

	if ( !(bbFrontSpan & bbPc(p,side, P) ) ) {
	   if ( !(bbFrontSpan & bbPc(p, Opp(side), P) ) ) {
		  AddMiscTwo(side, Data.rookOpenMg, Data.rookOpenEg);
		  if (bbFrontSpan & bbKingZone[side] ) *attCount += Data.rookOpenAttack;
	   }
	  else
	  {
		  AddMiscTwo(side, Data.rookSemiOpenMg, Data.rookSemiOpenEg);
		  if (bbFrontSpan & bbKingZone[side] ) *attCount += Data.rookSemiOpenAttack;
	  }
	}

	// evaluate rook on 7th rank if it does anything useful there
	if (SqBb(sq) & bbSeventh[side] ) {
       if ( bbPc(p, Opp(side), P) & bbSeventh[side]
	   || bbPc(p, Opp(side), K) & bbEighth[side]  
	   )  AddMiscTwo(side, Data.rookSeventhMg, Data.rookSeventhEg);
	}

    // check threats
	if (bbControl & kingStraightChecks[Opp(side)] )
		*attCount += ROOK_CAN_CHECK;

	// if we can attack enemy king from current square, test this possibility
	if ( bbRCanAttack[sq] [KingSq(p, side ^ 1) ]  
	&& ( bbControl & bbKingZone[side] ) ) {

       bbAttZone = bbControl & bbKingZone[side];
	
	   if (bbAttZone && p->pcCount[side][Q]) {
	      *attCount += Data.attMultiplier[R] * PopCntSparse(bbAttZone);
	      *woodCount += Data.woodCount[R]; 
	   }
	}
	
	AddMobility(R, side, PopCnt15(bbControl) );
	
	bbPieces &= bbPieces - 1;
  }
}

/*
Queen evaluation:

- real mobility
- safe contact checks
- check threats
- king attacks (also through other sliders)
*/

void sEvaluator::ScoreQ(sPosition *p, int *attCount, int *woodCount, int side) 
{
  int sq, contactSq;
  U64 bbControl, bbAttacks, bbContact, bbAttZone;
  U64 bbPieces      = bbPc(p, side, Q);
  U64 bbOccupied    = OccBb(p);
  U64 bbTransparent = bbPc(p, side, R) | bbPc(p, side, B);

  while (bbPieces) {
    sq = FirstOne(bbPieces);

	/* queen evaluation uses real mobility (without "transparent" pieces)
	/ because we want to detect queen contact checks */
	bbControl = GenCache.GetQueenMob(bbOccupied, sq);

	/**/ // contact checks - in the current form it's an even deal
	bbContact = bbControl & bbKingAttacks[ p->kingSquare[Opp(side)] ];
	while (bbContact) {
       contactSq = FirstOne(bbContact);
	   if ( Swap(p, sq, contactSq) >= 0 ) {
          *attCount += 30; 
		  break;
	   }
	   bbContact &= bbContact - 1;
	}
	/**/

    // check threats (TEST POPCNT VERSION)
	if (bbControl & ( kingStraightChecks[Opp(side)] | kingDiagChecks[Opp(side)] ) )
		*attCount += /*PopCntSparse( bbControl & ( kingStraightChecks[Opp(side)] | kingDiagChecks[Opp(side)] ) ) */ QUEEN_CAN_CHECK;

	/* If we can attack enemy king from current square, test this possibility;
	/ additionally, since queen is evaluated last, we know when an attack
	/ by it will not change score. */
    if (bbQCanAttack[sq] [KingSq(p, side ^ 1) ] 
	&& (*woodCount > 0 || p->pcCount[side][Q] > 1) )
	{
	   // consider also attacks through other pieces
	   if (bbControl && bbTransparent)
	      bbAttacks = RAttacks(bbOccupied ^ bbTransparent ^ SqBb(sq), sq);
	   else 
	      bbAttacks = bbControl;

	   bbAttZone = bbAttacks & bbKingZone[side];
	
	   if (bbAttZone) {
	      *attCount  += Data.attMultiplier[Q] * PopCntSparse(bbAttZone);
	      *woodCount += Data.woodCount[Q]; 
	   }
	}
	
	AddMobility(Q, side, PopCnt(bbControl) );
	    	
	bbPieces &= bbPieces - 1;
  }
}

void sEvaluator::AddMobility( int pc, int side, int cnt)
{
	mgMobility[side] += Data.mobBonusMg [pc] [cnt];
	egMobility[side] += Data.mobBonusEg [pc] [cnt];
}

void sEvaluator::AddMiscOne( int side, int val)
{
	mgMisc[side] += val;
	egMisc[side] += val;
}

void sEvaluator::AddMiscTwo( int side, int mg, int eg)
{
	mgMisc[side] += mg;
	egMisc[side] += eg;
}