/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "data.h"
#include "rodent.h"
#include "timer.h"
#include "trans.h"
#include "hist.h"
#include "search.h"
#include "parser.h"

void sParser::ReadLine(char *str, int n)
{
  char *ptr;

  if (fgets(str, n, stdin) == NULL)
    exit(0);
  if ((ptr = strchr(str, '\n')) != NULL)
    *ptr = '\0';
}

char *sParser::ParseToken(char *string, char *token)
{
  while (*string == ' ')
    string++;
  while (*string != ' ' && *string != '\0')
    *token++ = *string++;
  *token = '\0';
  return string;
}

void sParser::UciLoop(void)
{
  char command[4096], token[80], *ptr;
  sPosition p[1];

  setbuf(stdin, NULL);
  setbuf(stdout, NULL);
  SetPosition(p, START_POS);
  TransTable.Alloc(16);

  for (;;) {
    ReadLine(command, sizeof(command));
    ptr = ParseToken(command, token);
    if (strcmp(token, "uci") == 0) {
      PrintEngineHeader();
      PrintUciOptions(); 
	  printf("uciok\n");
    } else if (strcmp(token, "isready") == 0) {
      printf("readyok\n");
    } else if (strcmp(token, "setoption") == 0) {
      SetOption(ptr);
    } else if (strcmp(token, "position") == 0) {
      ParsePosition(p, ptr);
    } else if (strcmp(token, "ucinewgame") == 0) {
	  History.OnNewGame();
    } else if (strcmp(token, "go") == 0) {
      ParseGo(p, ptr);
    } else if (strcmp(token, "quit") == 0) {
      exit(0);
    }
  }
}

void sParser::SetOption(char *ptr)
{
  char token[80], name[80], value[80];

  ptr = ParseToken(ptr, token);
  name[0] = '\0';

  for (;;) {
    ptr = ParseToken(ptr, token);
    if (*token == '\0' || strcmp(token, "value") == 0)
      break;
    strcat(name, token);
    strcat(name, " ");
  }

  name[strlen(name) - 1] = '\0';

  if (strcmp(token, "value") == 0) {
    value[0] = '\0';

    for (;;) {
      ptr = ParseToken(ptr, token);
      if (*token == '\0')
        break;
      strcat(value, token);
      strcat(value, " ");
    }

    value[strlen(value) - 1] = '\0';
  }

  if (strcmp(name, "OwnMobility") == 0) {
	Data.ownMobility = atoi(value);
  } else if (strcmp(name, "OppMobility") == 0) {
	Data.oppMobility = atoi(value);
  } else if (strcmp(name, "OwnAttack") == 0) {
	Data.ownAttack = atoi(value);
  } else if (strcmp(name, "OppAttack") == 0) {
	Data.oppAttack = atoi(value);
  } else if (strcmp(name, "LMRbase") == 0) {
	Data.moveIsLate = atoi(value);
  } else if (strcmp(name, "LMRstep") == 0) {
	Data.lmrStep = atoi(value);
  } else  if (strcmp(name, "FutilityDepth") == 0) {
	Data.futilityDepth = atoi(value);
  } else  if (strcmp(name, "FutilityBase") == 0) {
	Data.futilityBase = atoi(value);
  } else if (strcmp(name, "FutilityStep") == 0) {
	Data.futilityStep = atoi(value);
#ifdef USE_WEAKENING
  } else if (strcmp(name, "UCI_Elo") == 0) {
	Data.elo = atoi(value);
#endif
  } else if (strcmp(name, "Hash") == 0) {
    TransTable.Alloc(atoi(value));
  } else if (strcmp(name, "Clear Hash") == 0) {
    TransTable.Clear();
  }
}

void sParser::PrintUciOptions() 
{
#ifdef USE_WEAKENING
	printf("option name UCI_Elo type spin default %d min 1000 max %d \n"    , Data.elo, MAX_ELO);
#endif
	printf("option name OwnMobility type spin default %d min 0 max 300\n"  , Data.ownMobility);
	printf("option name OppMobility type spin default %d min 0 max 300\n"  , Data.oppMobility);
	printf("option name OwnAttack type spin default %d min 0 max 300\n"    , Data.ownAttack);
	printf("option name OppAttack type spin default %d min 0 max 300\n"    , Data.oppAttack);
	printf("option name LMRbase type spin default %d min 1 max 100\n"      , Data.moveIsLate);
	printf("option name LMRstep type spin default %d min 0 max 100\n"      , Data.lmrStep);
	printf("option name FutilityDepth type spin default %d min 0 max 8\n"  , Data.futilityDepth);
	printf("option name FutilityBase type spin default %d min 0 max 1000\n", Data.futilityBase);
	printf("option name FutilityStep type spin default %d min 0 max 100\n" , Data.futilityStep);
    printf("option name Hash type spin default 16 min 1 max 4096\n");
    printf("option name Clear Hash type button\n");
}

void sParser::ParsePosition(sPosition *p, char *ptr)
{
  char token[80], fen[80];
  UNDO u[1];

  ptr = ParseToken(ptr, token);
  if (strcmp(token, "fen") == 0) {
    fen[0] = '\0';
    for (;;) {
      ptr = ParseToken(ptr, token);
      if (*token == '\0' || strcmp(token, "moves") == 0)
        break;
      strcat(fen, token);
      strcat(fen, " ");
    }
    SetPosition(p, fen);
  } else {
    ptr = ParseToken(ptr, token);
    SetPosition(p, START_POS);
  }
  if (strcmp(token, "moves") == 0)
    for (;;) {
      ptr = ParseToken(ptr, token);
      if (*token == '\0')
        break;
      DoMove(p, StrToMove(p, token), u);
      if (p->reversibleMoves == 0)
        p->head = 0;
    }
}

void sParser::ParseGo(sPosition *p, char *ptr)
{
  char token[80], bestmoveString[6];
  int pv[MAX_PLY];
  // TODO: move PV to search class

  Timer.Clear();
  pondering = 0;

  for (;;) {
    ptr = ParseToken(ptr, token);
    if (*token == '\0')
      break;
    if (strcmp(token, "ponder") == 0) {
      pondering = 1;
    } else if (strcmp(token, "wtime") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(W_TIME, atoi(token) );
    } else if (strcmp(token, "btime") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(B_TIME, atoi(token) );
    } else if (strcmp(token, "winc") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(W_INC, atoi(token) );
    } else if (strcmp(token, "binc") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(B_INC, atoi(token) );
    } else if (strcmp(token, "movestogo") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(MOVES_TO_GO, atoi(token) );
    } else if (strcmp(token, "depth") == 0) {
      ptr = ParseToken(ptr, token);
      Timer.SetData(MAX_DEPTH, atoi(token) );
	}
  }

  Timer.SetSideData(p->side);
  Timer.SetMoveTiming();
  
  Searcher.Think(p, pv);
  MoveToStr(pv[0], bestmoveString);
  if (pv[1]) {
    MoveToStr(pv[1], ponder_str);
    printf("bestmove %s ponder %s\n", bestmoveString, ponder_str);
  } else
    printf("bestmove %s\n", bestmoveString);
}

void sParser::PrintEngineHeader() 
{
    printf("id name Rodent 0.12\n");
    printf("id author Pawel Koziol (based on Sungorus by Pablo Vazquez)\n");
}