/*f
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "rodent.h"
#include "data.h"
#include "bitboard.h"
#include "bitmasks.h"
#include "move.h"
#include "selector.h"
#include "timer.h"
#include "trans.h"
#include "hist.h"
#include "parser.h"
#include "search.h"
#include "eval.h"

void sSearcher::Think(sPosition *p, int *pv)
{
  History.OnNewSearch();
  TransTable.ChangeDate();
  nodes           = 0;
  flagAbortSearch = 0;
  Timer.SetStartTime();
  Iterate(p, pv);
}

void sSearcher::Iterate(sPosition *p, int *pv) {
  int val = 0;
  int curVal;
  int aspiration;

  int localDepth = AdjustRootDepth(p); // get search depth (reducing it in case of one legal move)
  Timer.SetIterationTiming();          // define additional rules for starting next iteration
  Data.InitAsymmetric(p->side);        // set asymmetric eval parameters, dependent on the side to move
  Timer.SetData(FLAG_ROOT_FAIL_LOW, 0);

  for (rootDepth = ONE_PLY; rootDepth <= localDepth; rootDepth+=ONE_PLY) {

	 DisplayRootInfo();
	 aspiration = Data.aspiration; // TODO: progressive widening loop
    
	// first use aspiration window around the value from the last completed depth
	curVal = Search(p, 0, val-aspiration, val+aspiration, rootDepth, IS_PV, NO_NULL, 0, pv);
	if (flagAbortSearch) break;

	// if score is outside the window, re-search
	if (curVal >= val+aspiration || curVal <= val-aspiration) 
	{
        // fail-low, it might be prudent to assign some more time
        if (curVal <= val-aspiration) 
			Timer.SetData(FLAG_ROOT_FAIL_LOW, 1);

		curVal = Search(p, 0, val-3*aspiration, val+3*aspiration, rootDepth, IS_PV, NO_NULL, 0, pv);
        if (flagAbortSearch) break;

		// the second window
		if (curVal >= val+3*aspiration || curVal <= val-3*aspiration) 
            curVal = Search(p, 0, -INF, INF, rootDepth, IS_PV, NO_NULL, 0, pv);
        if (flagAbortSearch) break;
	}

	// abort root search if we don't expect to finish the next iteration
	if (Timer.FinishIteration() ) {
		DisplaySavedIterationTime();
		break; 
	}

	val = curVal;
  }
}

int sSearcher::Search(sPosition *p, int ply, int alpha, int beta, int depth, int isPvNode, int wasNull, int lastMove, int *pv)
{
  int best,                // best value found at this node
	  score,               // score returned by a search started in this node
	  move,                // a move we are searching right now
	  depthChange,         // extension/reduction value
	  newDepth,            // depth of a new search started in this node
	  newPv[MAX_PLY];      // new main line
  UNDO  undoData[1];       // data required to undo a move
  sSelector Selector;      // an object responsible for maintaining move list and picking moves

  // NODE INITIALIZATION
  int nullScore      = 0;  
  int nullRefutation = 0;  // a capture that refuted a null move
  int refutationSq   = NO_SQ; // target square of that capture
  int movesTried     = 0;  // count of moves that initiated new searches
  int flagIsReduced  = 0;  // are we in a reduced search? (guides re-searches)
  int flagCanPrune   = 0;  // can we statically prune in this node
  int flagInCheck    = InCheck(p); // are we in check at the beginning of the search?
  int flagMoveType;        // move type flag, supplied by NextMove()

#ifdef USE_WEAKENING
  int blunderCount   = 0;
#endif

  // at root we keep track whether a move has been found; if not, we let the engine
  // search for a bit longer, as this might indicate a fail-high/fail-low
  if (!ply && isPvNode) Timer.SetData(FLAG_NO_FIRST_MOVE, 1);

  // CHECK EXTENSION
  if (flagInCheck) {
	  depth += ONE_PLY;
	  // TODO: extend at low no. of replies (requires out of check / legal move generator)
  }
  
  // QUIESCENCE SEARCH ENTRY POINT
  if ( depth < ONE_PLY ) return Quiesce(p, ply, 0, alpha, beta, pv);

  nodes++;
  CheckInput();
  
  // EARLY EXIT CONDITIONS
  if ( flagAbortSearch )        return 0;
  if ( IsRepetition(p) && ply ) return 0;
  if ( DrawBy50Moves(p) )       return 0;
  if ( !flagInCheck && ply && RecognizeDraw(p) ) return 0;
  
  if (ply) *pv = 0;
  move = 0;

  // MATE DISTANCE PRUNING
   alpha = Max(-MATE+ply, alpha);
   beta  = Min( MATE-ply, beta);
   if (alpha >= beta)
        return alpha;

  // TRANSPOSITION TABLE READ
  // get transposition table sore or at least get a move for sorting purposes

  if (TransTable.Retrieve(p->hashKey, &move, &score, alpha, beta, depth, ply))
    return score;

  // INTERNAL ITERATIVE DEEPENING
  // in a pv node we try to get a hash move to order moves properly
  
  if (isPvNode && !move && depth >= 4*ONE_PLY && !flagInCheck ) {
	  Search(p, ply, alpha, beta, depth-2*ONE_PLY, IS_PV, NO_NULL, lastMove, newPv);
	  TransTable.Retrieve(p->hashKey, &move, &score, alpha, beta, depth, ply);
  }
  
  // safeguard against hitting max ply limit
  if (ply >= MAX_PLY - 1) return Eval.Return(p, alpha, beta);

  // STATIC NULL MOVE (idea taken from Stockfish)

  if (Data.useNull
  && !isPvNode
  && depth <= Data.minimalNullDepth
  && !wasNull
  && beta < MAX_EVAL
  && !flagInCheck
  && p->pieceMat[p->side] > 325
  ) {
    int delta = 150 + 20*depth;
	if (beta + delta <= Eval.ReturnFast(p) )
	   return beta + delta; 
    
  } // end of static null move code 

  // NULL MOVE
  // - we allow opponent to move twice in a row; if he cannot beat
  // beta this way, then we assume that there is no need to search any further.

  if (Data.useNull
  && !isPvNode
  && depth > Data.minimalNullDepth
  && !wasNull
  && beta < MAX_EVAL
  && !flagInCheck
  && p->pieceMat[p->side] > Data.matValue[B]
  ) {
    if ( beta <= Eval.Return(p, alpha, beta) ) 
	{
      DoNull(p, undoData);
	  newDepth = SetNullDepth(depth);
      nullScore = -Search(p, ply + 1, -beta, -beta + 1, newDepth, NO_PV, WAS_NULL, 0, newPv);

	  // Extract refutation of a null move from transposition table; 
	  // most of the time it will be capture and we will use this information
	  // to sort an evasion much higher than other quiet moves.
	  TransTable.Retrieve(p->hashKey, &nullRefutation, &score, alpha, beta, depth, ply);
	  
	  // get target square of null move refutation to sort escape moves a bit higher
	  if (nullRefutation != 0) refutationSq = Tsq(nullRefutation);
      UndoNull(p, undoData);

      if (flagAbortSearch) return 0; // timeout, "stop" command or mispredicted ponder move

	  // verify null move in the endgame
	  if (nullScore >= beta
	  && p->pieceMat[p->side] < 1600
	  && newDepth > 4 * ONE_PLY) {
      nullScore = Search(p, ply + 1, alpha, beta, newDepth-2*ONE_PLY, NO_PV, NO_NULL, lastMove, newPv);
	  }

      if (nullScore >= beta) {
        TransTable.Store(p->hashKey, 0, nullScore, LOWER, depth, ply);
		return Eval.Normalize(nullScore, MAX_EVAL); // checkmate from null move search isn't reliable
	  }
    }
  } // end of null move code 

  // SET FUTILITY PRUNING FLAG
  if ( Data.useFutility
  && depth < Data.futilityDepth * ONE_PLY  // we are sufficiently close to the leaf
  && beta < MAX_EVAL                       // we're not checkmating
  && alpha > -MAX_EVAL                     // we're not being checkmated
  && !isPvNode                             // we're not in a pv node
  && !flagInCheck                          // we're not in check
  ) {
   int nodeEval = TransTable.RefineScore( p->hashKey, Eval.ReturnFast(p) );

   // this node looks bad enough, so we may apply futility pruning
   if ( (nodeEval + SetFutilityMargin(depth) ) < beta ) flagCanPrune = 1;
  }

  // CREATE MOVE LIST AND START SEARCHING
  best = -INF;
  Selector.InitMoves(p, move, ply);

  // LOOP THROUGH THE MOVE LIST
  while ( (move = Selector.NextMove(refutationSq, &flagMoveType)) ) {

	DoMove(p, move, undoData);    
	
	if (IllegalPosition(p)) { 
		UndoMove(p, move, undoData); 
		continue; 
	}

#ifdef USE_WEAKENING // MAKE RANDOM BLUNDERS IN WEAKENING MODE
	if (Blunder(p, ply, depth, flagMoveType, move, lastMove)
	&& (movesTried > 1  || blunderCount < 2 ) ) {
		blunderCount++;
		UndoMove(p, move, undoData); 
		continue; 
	}
#endif

	movesTried++;                     // increase legal move count
	flagIsReduced  = 0;               // this move has not been reduced (yet)
	depthChange    = 0;               // no depth modification so far
	History.OnMoveTried(move);

	// EXTENSIONS might be placed here

	// FUTILITY PRUNING 
	if ( flagCanPrune                 // flag is set
	&& !depthChange                   // we have not extended this move
	&& !flagMoveType                  // we are considering ordinary move
	&& movesTried > 1                 // we have found at least one legal move
	) {
       UndoMove(p, move, undoData);
       continue;
	}

	// LATE MOVE PRUNING near the leaves
	if (Data.useLmr 
    &&  !isPvNode                      // we're not in a pv node
    &&  depth <= Data.minimalLmrDepth  // we are near the leaf
	&&  depthChange == 0               // it has not been extended
	&&  movesTried > 20                // it is sufficiently down the move list
	&&  !flagInCheck                   // we're not escaping a check
    &&  !InCheck(p)                    // we're not giving check	
	) {
       UndoMove(p, move, undoData);
       continue;
	}

	// LATE MOVE REDUCTION (lmr) - we reduce a move if:
	if (Data.useLmr 
    &&  !isPvNode                      // we're not in a pv node
    &&  depth > Data.minimalLmrDepth   // we have some depth left
	&&  depthChange == 0               // it has not been extended
	&&  movesTried > Data.moveIsLate   // it is sufficiently down the move list
	&&  !flagInCheck                   // we're not escaping a check
    &&  !InCheck(p)                    // we're not giving check
	&&  History.MoveIsBad(move)        // a move has bad history score
	&&  !flagMoveType                  // move is not special (capture, killer, hash move)
	) {             
		   depthChange -= SetLmrDepth(move,movesTried);
		   History.OnMoveReduced(move);
 	       flagIsReduced = 1;
	} // end of late move reduction code

	newDepth = depth - ONE_PLY + depthChange; // determine new depth

    re_search: // we come back here if a reduced search has not failed low

	// PRINCIPAL VARIATION SEARCH
	if (best == -INF )
      // while searching first move of a new node, isPvNode inherits parent node status
      score =   -Search(p, ply+1, -beta,    -alpha, newDepth, isPvNode, NO_NULL, move, newPv);
    else {
      score =   -Search(p, ply+1, -alpha-1, -alpha, newDepth, NO_PV, NO_NULL, move, newPv);
      if (!flagAbortSearch && score > alpha && score < beta)
        score = -Search(p, ply+1, -beta,    -alpha, newDepth, IS_PV, NO_NULL, move, newPv);
    }

    // FALLBACK TO NORMAL SEARCH IF REDUCED MOVE SEEMS GOOD
	if (Data.useReSearch && flagIsReduced && score > alpha) {
       flagIsReduced = 0;          // flag this search as unreduced         
       newDepth -= depthChange;    // undo reduction
	   goto re_search;
	}

    UndoMove(p, move, undoData);
	if (!ply && isPvNode) Timer.SetData(FLAG_NO_FIRST_MOVE, 0);
    
	if (flagAbortSearch) return 0; // timeout, "stop" command or mispredicted ponder move

    // BETA CUTOFF
	if (score >= beta) {
      History.OnGoodMove(p, move, depth / ONE_PLY, ply);
      TransTable.Store(p->hashKey, move, score, LOWER, depth, ply);
      return score;
    }

	// SCORE CHANGE
    if (score > best) {
      best = score;
      if (score > alpha) {
        alpha = score;
        if (isPvNode) BuildPv(pv, newPv, move);
		if (!ply) DisplayPv(score, pv);
      }
    }
  }

  // RETURN CORRECT CHECKMATE/STALEMATE SCORE
  if (best == -INF) return InCheck(p) ? -MATE + ply : 0;

  // SAVE SEARCH RESULT IN TRANSPOSITION TABLE
  if (*pv) {
	History.OnGoodMove(p, *pv, depth / ONE_PLY, ply);
    TransTable.Store(p->hashKey, *pv, best, EXACT, depth, ply);
  } else
    TransTable.Store(p->hashKey, 0, best, UPPER, depth, ply);

  return best;
}

// determine the amount a null move is reduced

int sSearcher::SetNullDepth(int depth) 
{
  int newDepth = depth - 3*ONE_PLY;
  if ( newDepth > 4*ONE_PLY ) 
	  newDepth -= ( ( (newDepth - 4*ONE_PLY) / 4) * ONE_PLY); 
  return newDepth;
}

int sSearcher::SetFutilityMargin(int depth) 
{
	return Data.futilityBase + depth * Data.futilityStep;
}

int sSearcher::SetLmrDepth(int move, int movesTried) 
{
    int depthChange = THREE_Q_PLY;

	// reductions for even later moves get fractionally bigger
	depthChange += QUARTER_PLY * (movesTried > Data.moveIsLate + Data.lmrStep);
	depthChange += QUARTER_PLY * (movesTried > Data.moveIsLate + 2*Data.lmrStep);
	depthChange += QUARTER_PLY * (movesTried > Data.moveIsLate + 3*Data.lmrStep);
	depthChange += QUARTER_PLY * (movesTried > Data.moveIsLate + 4*Data.lmrStep);

    return depthChange;
}

int sSearcher::IsRepetition(sPosition *p)
{
  int i;

  for (i = 4; i <= p->reversibleMoves; i += 2)
    if (p->hashKey == p->repetitionList[p->head - i])
      return 1;
  return 0;
}

int sSearcher::DrawBy50Moves(sPosition *p) 
{
    return( p->reversibleMoves > 100);
}

void sSearcher::CheckInput(void)
{
  char command[80];

  if (nodes & 4095 || rootDepth == ONE_PLY) return;

  if (InputAvailable()) {
    Parser.ReadLine(command, sizeof(command));
    if (strcmp(command, "stop") == 0)
      flagAbortSearch = 1;
    else if (strcmp(command, "ponderhit") == 0)
      pondering = 0;
  }
  
  // timeout
  if ( !pondering 
  && !Timer.IsInfiniteMode()
  && Timer.TimeHasElapsed()
  )
    flagAbortSearch = 1;
}

// Returns number of legal moves (in current form
// it is computationally expensive and used at root only)

int sSearcher::CountLegalMoves(sPosition *p, int cap)
{
  // TODO: ply as input variable
  int move;
  int moveCount = 0;
  int unusedFlag;
  sSelector Selector;
  UNDO  undoData[1];       // data required to undo a move

  Selector.InitMoves(p, 0, MAX_PLY); // prepare move selector

  while (( move = Selector.NextMove(0, &unusedFlag)) ) {

	DoMove(p, move, undoData);
    if (IllegalPosition(p)) { 
		UndoMove(p, move, undoData); 
		continue; 
	}
	moveCount++;
	UndoMove(p, move, undoData);
	if (moveCount >= cap) return moveCount;
  }
  return moveCount;
}

int sSearcher::AdjustRootDepth(sPosition *p)
{
  // get search depth
  int localDepth = Timer.GetData(MAX_DEPTH) * ONE_PLY;

  // reduce search depth if we have just one legal move
  // (we want to have some hash draft, so we search to depth of 4 plies)
  if (CountLegalMoves(p,3) < 2) localDepth = 4 * ONE_PLY;

  return localDepth;
}

int sSearcher::RecognizeDraw(sPosition *p) 
{
  // bare kings or Km vs K are classified as a draw and not searched any further
  if ( p->pcCount[WHITE][P] == 0 
  &&   p->pcCount[BLACK][P] == 0 ) 
  {		  
	  // Km vs K
	  if ( p->pieceMat[WHITE] + p->pieceMat[BLACK] < 400 ) {
	     if ( !IllegalPosition(p) ) return 1;
	  }
	  
      // Km vs Km, jut in case we take care that no king is on the rim 
	  if (p->pieceMat[WHITE] < 400 
	  &&  p->pieceMat[BLACK] < 400
	  &&  (bbPc(p, WHITE, K) & bbRim) == 0
	  &&  (bbPc(p, BLACK, K) & bbRim) == 0 ) {
		  if ( !IllegalPosition(p) ) return 1;
	  }
  } // no pawns

  return 0;
}

#ifdef USE_WEAKENING
int sSearcher::Blunder(sPosition *p, int ply, int depth, int flag, int move, int lastMove) {

	// we're playing the best game we can, so don't bother with weakening
	// (speed optimization)
	if (Data.elo == MAX_ELO || !Data.useWeakening ) return 0;
	
	// we try all root moves
	if (!ply) return 0;

	// stronger levels need more accuracy
	if (Data.elo > 1499 && ply < 3) return 0;
	if (Data.elo > 1999 && ply < 4) return 0;

    // we never miss a queen promotion
	if ( MoveType(move) == Q_PROM) return 0;

	// we always try to capture the last piece that moved
	if (Tsq(move) == Tsq(lastMove) ) return 0;

	// we never miss captures by a pawn
	if ( p->pc[Tsq(move)] == P 
	&& File(Tsq(move)) != File(Fsq(move)) ) return 0; 

	// we never miss short captures
	if ( Data.distance[Fsq(move)][Tsq(move)] > 9
	&& flag > FLAG_KILLER_MOVE 
	&& flag != FLAG_HASH_MOVE ) return 0;

	int blunderRate = rand() % Data.elo;
	
	// calculate blunder probability (less blunders near the root in deeper search)
	if ( blunderRate + depth < MAX_ELO - Data.elo) return 1;
	
	return 0;
}
#endif